#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/extensions/shape.h>
#include <X11/extensions/XTest.h>
#include <X11/extensions/XShm.h>
#include <Imlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

Display *disp;
Window win;
ImlibData *id;

char root = 0;
char scale = 0;
char scale_w = 0;
char scale_h = 0;
char integer_scale = 0;
char keep_aspect = 0;
char center_h = 0;
char center_v = 0;
int  x = 0, y = 0, w = 0, h = 0;
int  ax = -9999, ay = -9999, aw = 0, ah = 0;
char *file = NULL;

void
print_usage(void)
{
  printf("Usage:\n"
	 "xsri [options] image_file\n"
	 "\n"
	 "where options are any of:\n"
	 "  -geometry [WIDTHxHEIGHT][+X+Y]\n"
	 "  -root\n"
	 "  -scale\n"
	 "  -scale-width\n"
	 "  -scale-height\n"
	 "  -avoid [WIDTHxHEIGHT][+X+Y]\n"
	 "  -center-horizontal\n"
	 "  -center-vertical\n"
	 "  -integer-scale\n"
	 "  -keep-aspect\n");
  exit(1);
}

void
parse_args(int argc, char **argv)
{
  int i;
  
  for (i = 1; i < argc; i++)
    {
      if ((i < (argc - 1)) &&
	  ((!strcmp(argv[i], "-geometry")) ||
	  (!strcmp(argv[i], "-g"))))
	{
	  i++;
	  XParseGeometry(argv[i], &x, &y, &w, &h);
	}
      else if ((i < (argc - 1)) &&
	  ((!strcmp(argv[i], "-avoid"))))
	{
	  i++;
	  XParseGeometry(argv[i], &ax, &ay, &aw, &ah);
	}
      else if ((!strcmp(argv[i], "-help")) ||
	       (!strcmp(argv[i], "--help")) ||
	       (!strcmp(argv[i], "-h")))
	print_usage();
      else if (!strcmp(argv[i], "-root"))
	root = 1;
      else if (!strcmp(argv[i], "-scale"))
	scale = 1;
      else if (!strcmp(argv[i], "-scale-width"))
	scale_w = 1;
      else if (!strcmp(argv[i], "-scale-height"))
	scale_h = 1;
      else if (!strcmp(argv[i], "-integer-scale"))
	integer_scale = 1;
      else if (!strcmp(argv[i], "-keep-aspect"))
	keep_aspect = 1;
      else if (!strcmp(argv[i], "-center-horizontal"))
	center_h = 1;
      else if (!strcmp(argv[i], "-center-vertical"))
	center_v = 1;
      else 
	file = argv[i];
    }
}

ImlibImage *
load_image(void)
{
  ImlibImage *im = NULL;
  
  im = Imlib_load_image(id, file);
  if (!im)
    {
      fprintf(stderr, "file %s canot be loaded\n", file);
      exit(1);
    }
  return im;
}

int
intersetcs_quad(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2)
{
  if (((x1 < (x2 + w2)) && ((x1 + w1) > x2)) &&
      ((y1 < (y2 + h2)) && ((y1 + h1) > y2)))
    {
      if ((x1 + (w1 / 2)) < (x2 + (w2 / 2)))
	{
	  if ((y1 + (h1 / 2)) < (y2 + (h2 / 2)))
	    return 1;
	  else
	    return 4;
	}
      else
	{
	  if ((y1 + (h1 / 2)) < (y2 + (h2 / 2)))
	    return 2;
	  else
	    return 3;
	}
    }
  return 0;
}

void
calc_geometry(ImlibImage *im)
{
  int num, quad, relx = 0, rely = 0;

  if (x < 0)
    relx = 1;
  if (y < 0)
    rely = 1;
  if (w == 0)
    w = im->rgb_width;
  if (h == 0)
    h = im->rgb_height;
  if (scale_w)
    w = DisplayWidth(disp, DefaultScreen(disp));
  if (scale_h)
    h = DisplayHeight(disp, DefaultScreen(disp));
  if (scale)
    {
      w = DisplayWidth(disp, DefaultScreen(disp));
      h = DisplayHeight(disp, DefaultScreen(disp));
    }
  if (keep_aspect)
    {
      if (((w < 10) / im->rgb_width) < ((h < 10) / im->rgb_height))
	h = (w * im->rgb_height) / im->rgb_width;
      else
	w = (h * im->rgb_width) / im->rgb_height;
    }
  if (integer_scale)
    {
      if (w < im->rgb_width)
	{
	  num = im->rgb_width / w;
	  w = im->rgb_width / num;
	}
      else
	{
	  num = h / im->rgb_height;
	  w = num * im->rgb_height;
	}
      if (h < im->rgb_height)
	{
	  num = im->rgb_height / h;
	  h = im->rgb_height / num;
	}
      else
	{
	  num = h / im->rgb_height;
	  h = num * im->rgb_height;
	}
    }
  if (aw > 0)
    {
      if (ax <= -9999)
	{
	  ax = (DisplayWidth(disp, DefaultScreen(disp)) - aw) / 2;
	  ay = (DisplayHeight(disp, DefaultScreen(disp)) - ah) / 2;
	}
      quad = intersetcs_quad(x, y, w, h, ax, ay, aw, ah);
      switch (quad)
	{
	 case 1:
	 case 2:
	  y = ay - h;
	  if (y < 0)
	    {
	      y = 0;
	      h = ay;
	    }
	  break;
	 case 3:
	 case 4:
	  y = ay + ah;
	  if ((y + h) > DisplayHeight(disp, DefaultScreen(disp)))
	    {
	      h = DisplayHeight(disp, DefaultScreen(disp)) - (ay + ah);
	      y = ay + ah;
	    }
	  break;
	 default:
	  break;
	}
    }
  if (keep_aspect)
    {
      if (((w < 10) / im->rgb_width) < ((h < 10) / im->rgb_height))
	h = (w * im->rgb_height) / im->rgb_width;
      else
	w = (h * im->rgb_width) / im->rgb_height;
    }
  if (integer_scale)
    {
      if (w < im->rgb_width)
	{
	  num = im->rgb_width / w;
	  if ((num * w) < im->rgb_width)
	    num++;
	  w = im->rgb_width / num;
	}
      else
	{
	  num = h / im->rgb_height;
	  w = num * im->rgb_height;
	}
      if (h < im->rgb_height)
	{
	  num = im->rgb_height / h;
	  if ((num * h) < im->rgb_height)
	    num++;
	  h = im->rgb_height / num;
	}
      else
	{
	  num = h / im->rgb_height;
	  h = num * im->rgb_height;
	}
    }
  if (relx)
    x += DisplayWidth(disp, DefaultScreen(disp)) - w;
  if (rely)
    y += DisplayHeight(disp, DefaultScreen(disp)) - h;
  if (center_h)
    x = (DisplayWidth(disp, DefaultScreen(disp)) - w) / 2;
  if (center_v)
    y = (DisplayHeight(disp, DefaultScreen(disp)) - h) / 2;
}

void
display_image_window(ImlibImage *im)
{
  Imlib_paste_image(id, im, DefaultRootWindow(disp), x, y, w, h);
  Imlib_kill_image(id, im);
  XSync(disp,False);
}

void
display_root(ImlibImage *im)
{
  Pixmap pmap;
  
  Imlib_render(id, im, w, h);
  if ((x == 0) && (y == 0))
    pmap = Imlib_move_image(id, im);
  else
    {
      Pixmap pmap2;
      GC gc;
      XGCValues gcv;
      
      pmap2 = Imlib_move_image(id, im);
      pmap = XCreatePixmap(disp, DefaultRootWindow(disp), w, h, id->x.depth);
      gcv.tile = pmap2;
      gcv.ts_x_origin = x % w;
      gcv.ts_y_origin = y % h;
      gcv.fill_style = FillTiled;
      gc = XCreateGC(disp, pmap2, 
		     GCTile | GCTileStipXOrigin | GCTileStipYOrigin, &gcv);
      XFillRectangle(disp, pmap2, gc, 0, 0, w, h);
      XFreeGC(disp, gc);
      Imlib_free_pixmap(id, pmap2);
    }
  XSetWindowBackgroundPixmap(disp, DefaultRootWindow(disp), pmap);
  XClearWindow(disp, DefaultRootWindow(disp));
  Imlib_free_pixmap(id, pmap);
  Imlib_kill_image(id, im);  
  XSync(disp,False);
}

int 
main(int argc, char **argv)
{
  ImlibImage *im;

  parse_args(argc, argv);
  if (!file)
    print_usage();
  disp=XOpenDisplay(NULL);
  id=Imlib_init(disp);
  im = load_image();
  calc_geometry(im);
  if (root)
    display_root(im);
  else
    display_image_window(im);
  return 0;
}
